# ----------------------------------------------------------------------------
# 
# TITLE - util
# AUTHOR - Stephanie Monty
# PROJECT - MAVISIM
# 
# ----------------------------------------------------------------------------

### Imports
# Standard
import numpy as np
# Astropy
from astropy.table import Table, Column
from astropy import units as u
# Scipy
from scipy import interpolate

def input_coo(input_par, source):
	""" Take input parameters and parse them as an astropy table for comparison with DAOphot.

	Args:
		input_par: input parameters, e.g., from `input_parameters.py`.
		source (`Source` object): the `Source` object containing all of the sources simulated.
		
	Returns:
		trimmed_cat = an astropy table containing the TRUE input positions (static distortion, 
		sub-pixel positioning and proper motion taken into account) and additional information necessary to compare with the DAOPhot output.
	"""
	# Note we need to take two different coordinate conventions into account here, the convention of the CCD (x -> y) and the convention
	# of DAOPhot (starts at position 1, 1)

	# Find the full FoV in pixels, this contains the buffer (which is trimmed later) to contain light from stars outside the FoV
	full_fov_pix = int(input_par.MAVIS_fov + input_par.buffer)/input_par.ccd_sampling

	# Convert the position to pixels (remove the knowledge of the static distortion and add the proper motion (if any))
	x_pos = np.array(np.around(((source.x_pos/input_par.ccd_sampling) + full_fov_pix/2.0), 0), int)
	true_x = x_pos + source.x_dist - source.static_dist[:, 0] + source.x_pm + 1

	y_pos = np.array(np.around(((source.y_pos/input_par.ccd_sampling) + full_fov_pix/2.0), 0), int)
	true_y = y_pos + source.y_dist - source.static_dist[:, 1] + source.y_pm + 1

	# Create the final table
	input_coo = Table(data = (source.star,),names=("Star",))
	input_coo.add_column(Column(source.flux),name="Flux")
	input_coo.add_column(Column(source.ra),name="RA")
	input_coo.add_column(Column(source.dec),name="Dec")
	input_coo.add_column(Column(true_x), name="CCD_Mapped_X")
	input_coo.add_column(Column(source.x_pm), name="CCD_Mapped_PM_X")
	input_coo.add_column(Column(source.static_dist[:, 0]), name="X Static Dist")
	input_coo.add_column(Column(true_y), name="CCD_Mapped_Y")
	input_coo.add_column(Column(source.y_pm), name="CCD_Mapped_PM_Y")
	input_coo.add_column(Column(source.static_dist[:, 1]), name="Y Static Dist")

	# Roughly trim the input catalogue to only include stars in the MAVIS FoV
	# Set the boundary to include stars just outside of the CCD limit (diagonal)
	image_size = (input_par.MAVIS_fov + input_par.buffer)/input_par.ccd_sampling

	r = np.sqrt(2 * (input_par.ccd_size/2.0)**2) + 100

	r_coo = np.sqrt((int(image_size/2.0) - input_coo["CCD_Mapped_X"])**2 + 
					(int(image_size /2.0) - input_coo["CCD_Mapped_Y"])**2)

	trim = np.where(r_coo <= r)[0]
	trimmed_cat = input_coo[trim]

	# Correct for the change from the larger buffered FoV to the FoV of the MAVIS CCD
	trimmed_cat["CCD_Mapped_X"] = trimmed_cat["CCD_Mapped_X"] - (image_size/2 - input_par.ccd_size/2)
	trimmed_cat["CCD_Mapped_Y"] = trimmed_cat["CCD_Mapped_Y"] - (image_size/2 - input_par.ccd_size/2)

	return trimmed_cat

def add_all_noise(input_par, image, exp_time):
    """ Adds all noise effects onto noiseless image.

    Args:
        input_par: input parameters object.
        image (`np.ndarray`): 2D noise-free image.
        exp_time (`float`): total exposure time in seconds.
        
    Returns:
        image_adu (`np.ndarray`): final image + noise in ADU
    """

    # Create the true image by randomly sampling Poissons with expectation intervals = N* + Nsky
    # Add the readnoise on top by randomly sampling Gaussians with a sigma = readnoise
    
    if input_par.noise_mode == "slow":
        # Read noise in photons
        read_noise = input_par.slow_rdnoise
        
    if input_par.noise_mode == "fast":
        # Read noise in photons
        read_noise = input_par.fast_rdnoise

    # Get the constant sky value in photons/pixel
    sky_value = add_constant_sky_pixel(input_par, exp_time)

    # Remove the chance of negative N photons (only sky in that region), this is due to artifacts from the PSF creation
    image_addsky = image + sky_value
    image_addsky[np.where(image_addsky  < 0)] = sky_value

    # Account for the difference in photons between M1 and the imager
    image_addsky_mirror = image_addsky * input_par.AOMthruput * input_par.VLTthruput

    # Add the shot noise (from Poissonian stats)
    # We are not sky limited on the pixel scale - Poisson statistics
    rng = np.random.default_rng()

    # Draw the image in photons + photon noise
    true_im = rng.poisson(lam=image_addsky_mirror)

    # Convert from photons to electrons
    im_elec = true_im  * input_par.QE

    # Add Read Noise
    im_addnoise = im_elec + np.random.normal(loc=np.zeros([true_im.shape[0], true_im.shape[0]]), scale=read_noise)

    # Convert the final image from electrons to ADU
    # Cap at the saturation point
    if input_par.sat_point  == 0:
        sat_point = np.int(2**input_par.bit_depth - 1) * input_par.gain

    else:
        sat_point = input_par.sat_point

    # Convert from electrons to ADU, round to integers to be physical
    image_adu = (im_addnoise * input_par.gain).astype(np.int)
    image_adu[image_adu > sat_point] = sat_point
	
    return (image_adu)

def add_constant_sky_pixel(input_par, exp_time):
	""" Add constant-valued sky background.
	Args:
        input_par: input parameters object.    
		exp_time (`float`): exposure time in seconds to convert from photons/s to photons.
		
	Returns:
		sky_value (`float`): a global sky value in photons to add to every pixel
	""" 

	square_arcsec_pervoxel = (input_par.ccd_sampling**2) * u.arcsec**2

	# Assuming the surface brightness is passed in as mag/arcsec^2 do the following:
	flux_jy = (3631 * 10**(-1*input_par.surf_bright/2.5))*u.Jy*u.nm**(-1)
	flux_ph_s_nm_cm2 = flux_jy * (1.51e3/input_par.psf_wavelength)
	flux_ph_s_nm_m2 = flux_ph_s_nm_cm2 * 10**4
	flux_ph_s_m2 = flux_ph_s_nm_m2 * input_par.filt_width #in nm
	sky_value = flux_ph_s_m2 * square_arcsec_pervoxel  * exp_time * input_par.collecting_area

	return sky_value.value

def make_static_dist_map(input_par):
    """ Make static distortion map from loaded distortion samples in `input_par`.

    Args:
        input_par: input parameters object.    

    Returns:
        dist_x_func_degmm:  
        dist_y_func_degmm:  
    """
    # Field_y(deg) Hx Hy  Predicted_x(mm)  Predicted_y(mm)  Real_x(mm)  Real_y(mm)

    field_x = input_par.static_distort["Field_x(deg)"]
    field_y = input_par.static_distort["Field_y(deg)"]

    dist_x = input_par.dynamic_amp * (input_par.static_distort["Predicted_x(mm)"] - input_par.static_distort["Real_x(mm)"])
    dist_y = input_par.dynamic_amp * (input_par.static_distort["Predicted_y(mm)"] - input_par.static_distort["Real_y(mm)"])

    # Create an array of the field positions and the distortion at each pt (the difference)
    dist_all = np.empty([dist_x.shape[0], 4])
    dist_all[:, 0] = field_x
    dist_all[:, 1] = field_y
    dist_all[:, 2] = dist_x
    dist_all[:, 3] = dist_y

    grid_vals = np.unique(field_x)

    # Create grids to save the distortion (difference) in x and y at each point in the grid
    # This is necessary for the interpolation of the distortion (grid-wise interpolation)
    dist_x_grid = np.zeros([len(grid_vals),len(grid_vals)])
    dist_y_grid = np.zeros([len(grid_vals),len(grid_vals)])

    num = 0

    for x in grid_vals:
        sub_array = dist_all[np.where(dist_all[:, 0] == x), :][0]

        for row in np.arange(0, sub_array.shape[0]):

            dist_x_grid[num, row] = sub_array[row, 2]
            dist_y_grid[num, row] = sub_array[row, 3]

        num+=1

    dist_x_func_degmm = interpolate.RectBivariateSpline(grid_vals, grid_vals, dist_x_grid)
    dist_y_func_degmm = interpolate.RectBivariateSpline(grid_vals, grid_vals, dist_y_grid)
    
    return (dist_x_func_degmm, dist_y_func_degmm)


# ----------------------------------------------------------------------------
#
# TITLE - Source
# AUTHOR - Stephanie Monty
# PROJECT - MAVISIM
# CONTENTS:
#	V1.1 DIFFERENCES:
# 	1. NOT possible to add NGS tip-tilt jitter term on the fly, the NGS constellation is fixed as the best-case scenario (three stars, spaced evenly throughout the field)
#      and is already built into the e2e PSFs.
#
#	1. NOTE: Very important, this program and all programs assume all projection effects have been taken into account, the cos(Dec) term has been
#			 dealt with
#
#   1. This class creates an object with the required information to create a simulated image. This information is accessible via object attributes:
#		         star  =  Number assigned to the star to help with tracking through epochs in PM sims
#		         flux  =  Monochromatic flux of the star (photons/s) passed to the program by the input data
#		       ra/dec  =  RA and Dec of input star to aid in velocity dispersion calculations
#		          x/y  =  Distance from the centre of the field (in arcseconds)
#		    x_pm/y_pm  =  Proper motion of the star (mas)
#		x_dist/y_dist  =  Sub-pixel shifts due to a) the star falling at at a non-integer pixel position, b) static distortion present
#		    gauss_pos  =  Position of the MV Gaussian 
#		    gauss_cov  =  Covariance of the MV Gaussian 
#		  static_dist  =  Static distortion applied [X,Y]
#
#
# ----------------------------------------------------------------------------

### Docstrings and metadata:
'''
Generates a source object containing all the point sources used to create the final image. 
The info necessary to build a gaussian is created for each point source taking into account sub-pixel shifts due to distortion and non-integer pixel positions.

'''
__author__ = "Stephanie Monty"

### Imports

# Standard
import numpy as np
from tqdm import tqdm

# Project-specific
from mavisim import make_static_dist_map

class Source:
	"""
	The source object allows access to the parameters of each source (star) required
	to compute the MAVISIM image.

	Args:
		input_par : imported input parameter python file, e.g., `input_parameters.py`.
		exp_time (`float`): exposure time in seconds to simulate.
		static_dist (`bool`, optional): if `True`, add static distortion defined in `input_par`.
		stat_amp (`float`, optional): scaling factor to apply to distortions.
		tt_amp (`float`, optional): scaling factor to apply to tip-tilt blurring kernel width.
	
	Attributes:
		exp_time (`float`): exposure time in seconds to simulate.
		star (`np.ndarray` of `int`): unique ID of each star.
		flux (`np.ndarray` of `float`): flux of each star.
		ra (`np.ndarray` of `float`): RA of each star.
		dec (`np.ndarray` of `float`): Dec of each star.
		x_pos (`np.ndarray` of `float`): X position of each star (in arcsec).
		x_pm (`np.ndarray` of `float`): X proper motion of each star (in pixels).
		x_dist (`np.ndarray` of `float`): sub-pixel X position shift of each star (in pixels).
		y_pos (`np.ndarray` of `float`): Y position of each star (in arcsec).
		y_pm (`np.ndarray` of `float`): Y proper motion of each star (in mas/year).
		y_dist (`np.ndarray` of `float`): sub-pixel Y position shift of each star (in pixels).
		gauss_pos (`np.ndarray` of `float`): X/Y-position of each star (in arcsec).
		gauss_cov (`np.ndarray` of `float`): covariance of Gaussian kernel to simulate tip-tilt blurring (in arcsec^2).
		static_dist	(`np.ndarray` of `float`): static distortion to apply to each source.
	"""

	def __init__(self, input_par, exp_time, static_dist=False, stat_amp=1.0, tt_amp=1.0, use_cov=False):
		"""Create a MAVISIM Source data object
		"""
		self._input_par = input_par
		self.exp_time = exp_time
		self._static_dist_flag = static_dist
		self._stat_amp = stat_amp
		self._tt_amp = tt_amp
		# Add static field distortion to the image?, If so, create the functions necessary to determine the distortion to add
		if self._static_dist_flag == True:
			(self._dist_x_func_degmm, self._dist_y_func_degmm) = make_static_dist_map(self._input_par)
		else:
			self._dist_x_func_degmm = self._dist_y_func_degmm = None

		# Determine the covariance matrix (identical for all the stars) if it is desired
		self.cov_mat = self._find_covmat() if use_cov else None

	def build_source(self):
		""" From the data stored in the object, compute the source data as required
		"""
		dtype = np.float64
		num_stars   = self._input_par.input_cat["X"].shape[0]
		self.star   = np.zeros([num_stars],dtype=int)
		self.flux   = np.zeros([num_stars],dtype=dtype)
		self.ra     = np.zeros([num_stars],dtype=dtype)
		self.dec    = np.zeros([num_stars],dtype=dtype)
		self.x_pos  = np.zeros([num_stars],dtype=dtype)
		self.x_pm   = np.zeros([num_stars],dtype=dtype)
		self.x_dist = np.zeros([num_stars],dtype=dtype)
		self.y_pos  = np.zeros([num_stars],dtype=dtype)
		self.y_pm   = np.zeros([num_stars],dtype=dtype)
		self.y_dist = np.zeros([num_stars],dtype=dtype)
		self.gauss_pos   = np.zeros([num_stars,2],dtype=dtype)
		self.gauss_cov   = np.zeros([num_stars,2,2],dtype=dtype)
		self.static_dist = np.zeros([num_stars,2],dtype=dtype)
		
		for star in tqdm(range(num_stars)):
			star_info = self._input_par.input_cat[star]
			self._compute_row(star,star_info)

	def _compute_row(self, row, star_info):
		"""
		Args:
			star_info = row of the astropy table containing all the info required for that star
			
		Returns:
			row_info = the relevant info formatted into a row to return and append the final table with
		"""

		# Save the sub-pixel shift that comes from converting 15.1 degrees to pixels, keep track of this to shift later
		(x_pos_dist, y_pos_dist, static_dist) = self._find_shift(star_info)
		
		# PM in arcseconds
		pm_x = star_info["PM_X"]
		pm_y = star_info["PM_Y"]
		
		# Multiply the input flux by the exposure time
		total_flux = star_info["Flux"] * self.exp_time

		# Create a column vector for the x,y coordinates of the centre of the Gauss
		# All inputs are in arcseconds, convention is centre of FoV at (0, 0)
		gauss_mux = star_info["X"] + ((self._input_par.gauss_offset + x_pos_dist) * self._input_par.ccd_sampling) + pm_x
		gauss_muy = star_info["Y"] + ((self._input_par.gauss_offset + y_pos_dist) * self._input_par.ccd_sampling) + pm_y
		gauss_mu = np.r_[gauss_mux, gauss_muy]
		
		# Return the formatted info to append as a row to the final astropy table
		# Swap the x and y for the CCD convention

		self.star[row]   = star_info["Star"]
		self.flux[row]   = total_flux
		self.ra[row]     = star_info["RA"]
		self.dec[row]    = star_info["Dec"]
		self.x_pos[row]  = star_info["X"]
		self.x_pm[row]   = pm_x
		self.x_dist[row] = x_pos_dist
		self.y_pos[row]  = star_info["Y"]
		self.y_pm[row]   = pm_y
		self.y_dist[row] = y_pos_dist
		self.gauss_pos[row]   = gauss_mu
		self.gauss_cov[row]   = self.cov_mat
		self.static_dist[row] = static_dist

	def _find_shift(self, star_info):
		"""
		Args:
			star_info = row of the astropy table containing all the info required for that star
			
		Returns:
			x/y_dist = the sub-pixel shift in the x/y-direction for the star

		"""
		# TODO: this can be made significantly cleaner
		if self._static_dist_flag == True:
			# Retrieve the interpolated distortion at each point and convert from the default output (mm) to pixels
			# Input to functions is in degrees
			x_deg = star_info["X"]/3600.0
			y_deg = star_info["Y"]/3600.0

			# Static distortion for the star given the location, converting from mm to pixels
			x_static_dist = (((self._dist_x_func_degmm(x_deg, y_deg) * self._input_par.plate_scale)/self._input_par.ccd_sampling)[0][0]) * self._stat_amp
			y_static_dist = (((self._dist_y_func_degmm(x_deg, y_deg) * self._input_par.plate_scale)/self._input_par.ccd_sampling)[0][0]) * self._stat_amp

			# Sum the two sub-pixel terms to find the final distortion
			final_x_dist = x_static_dist
			final_y_dist = y_static_dist

			return (final_x_dist, final_y_dist, np.asarray([x_static_dist, y_static_dist]))

		else:
			# Sub-pixel shift is solely associated with the conversion from arcseconds to pixels
			final_x_dist = 0.0
			final_y_dist = 0.0

			return (final_x_dist, final_y_dist, np.asarray([0, 0]))

	def decimate(self, nstar):
		"""Decimate the list of objects in the object (e.g., for faster simualtions).

		Args:
			nstar (`int`): number of stars to reduce list down to. The resulting object will
			only contain the first `nstar` stars.
		"""
		if nstar >= self.star.shape[0]:
			raise ValueError(f"Already less than {nstar:d} stars in catalogue ({self.star.shape[0]:d})")
		self.star        =  self.star[:nstar]
		self.flux        =  self.flux[:nstar]
		self.ra          =  self.ra[:nstar]
		self.dec         =  self.dec[:nstar]
		self.x_pos       =  self.x_pos[:nstar]
		self.x_pm        =  self.x_pm[:nstar]
		self.x_dist      =  self.x_dist[:nstar]
		self.y_pos       =  self.y_pos[:nstar]
		self.y_pm        =  self.y_pm[:nstar]
		self.y_dist      =  self.y_dist[:nstar]
		self.gauss_pos   =  self.gauss_pos[:nstar]
		self.static_dist =  self.static_dist[:nstar]

	def _find_covmat(self):
		"""
		At the moment this only needs to be called once, as all the stars have the same covariance matrix.
		This is for the case that the TT residual is already captured in the e2e PSFs

		Args:
			star_info = row of the astropy table containing all the info required for that star
			
		Returns:
			cov_mat = covariance matrix to create the TT kernel

		"""

		# For now the semi-major and semi-minor axes will be identical (assuming only vibration and C.D. terms) 
		# Leave the option to add an additional skewed TT residual kernel in the future
		ttres_semi_major = 0
		ttres_semi_minor = 0

		# Add the fixed vibration and CD terms
		semi_major = np.sqrt(ttres_semi_major**2 + self._input_par.cd_term**2 + self._input_par.vib_term**2) * self._tt_amp
		semi_minor = np.sqrt(ttres_semi_minor**2 + self._input_par.cd_term**2 + self._input_par.vib_term**2) * self._tt_amp
		
		# Convert to arcseconds
		semi_major_arcsec = semi_major * self._input_par.ccd_sampling
		semi_minor_arcsec = semi_minor * self._input_par.ccd_sampling

		# For some reason the array is not oriented the correct way so y and x have to be swapped
		theta = np.pi/2

		# Correct the rotation about the first diagonal by swapping the sine and cosine of the angles
		c, s = np.sin(theta), np.cos(theta)
		R = np.array([[c, -s], [s, c]])
		RT = np.array([[c, s], [-s, c]])
		S = np.array([[semi_major_arcsec**2, 0], [0, semi_minor_arcsec**2]])

		# Generate the covariance matrix following a rotation
		# See https://www.visiondummy.com/2014/04/geometric-interpretation-covariance-matrix/
		cov_mat = R @ S @ RT
		
		return (cov_mat)


import numpy as np
from astropy.io import fits
from tqdm import tqdm

class PSF:
    """PSF object class.

    Helper class to handle PSF data.
    
    Args:
        fits_ext (`astropy.io.fits`): opened fits file containing PSF data and required
        header data (XPOS,YPOS,LAMBDA).
        padto (`list` of `int`): shape of desired Fourier transform array.
        dtype (`np.dtype`, optional): desired complex dtype of Fourier array.

    Attributes:
        fft_data (`np.ndarray`): array storing the (minimal) rfft2 data of the given PSF.
        xpos (`float`): x-position of PSF point source in arcsec.
        ypos (`float`): y-position of PSF point source in arcsec.
        Lambda (`float`): wavelength used to capture the PSF.
    """

    def __init__(self, fits_ext, padto, *, dtype=np.complex128):
        """Init PSF and perform RFFT2
        """
        self.fft_data = (np.fft.rfft2(fits_ext.data/fits_ext.data.sum(),s=padto)).astype(dtype)
        self.xpos     = float(fits_ext.header["YPOS"])
        self.ypos     = float(fits_ext.header["XPOS"])
        self.Lambda   = float(fits_ext.header["LAMBDA"])
    
class TileGenerator:
    """
    Object for generating tiles to be sliced into final image.

    Args:
        source (`Source` object): Object containing all of the source data, as defined in `Source.py`.
        psfs_file (`str`): path to `.fits` file containing all PSFs and metadata.
        gauss_width_pix (`int`): support size in pixels to build the Gaussian star kernel.
        dtype (`np.dtype`, optional): complex data type to work with in DFT space.
        which_psf (`int`, optional): If specified, use fits HDU[which_psf+1] PSF only.

    Attributes:
        source (`Source` object): collection of sources in `Source`-type object.
        psfs (`list` of `PSF` objects): list of PSF objects (see PSF class).
        pixsize (`float`): pixel size in arcsec used to build tile.
        gauss_width_pix (`int`): width of Gaussian square support in pixels
        static (`bool`) : if `True`, use a static PSF, otherwise use field variable PSF.
    """
    
    def __init__( self, source, psfs_file, gauss_width_pix, *,
             dtype = np.complex128, which_psf = None):
        """Initialise tile generator object by preparing source list and PSFs
        """
        
        self.dtype=dtype

        # Parse source info:
        self.source_flux = source.flux.copy()
        self.source_pos  = source.gauss_pos.copy()
        
        # covariance matrix of gaussian kernel (default None)
        if source.cov_mat is not None:
            self.cov_mat  = source.cov_mat.copy().astype(
                np.float32 if self.dtype==np.complex64 else np.float64
            )
        else:
            self.cov_mat = None
        
        # Define nominal image geometry:
        # TODO read this from fits Primary HDU metadata
        self.pixsize          = 3.75e-3         
        self.gauss_width_pix  = gauss_width_pix
        self.gauss_width_as   = self.gauss_width_pix*self.pixsize
        self.gauss_dim        = np.array([self.gauss_width_pix,self.gauss_width_pix])

        # Load PSF file and define Fourier Image Geometry:
        self.psf_file          = psfs_file
        psfs_fits              = fits.open(psfs_file, lazy_load_hdus=True)
        # TODO read this from fits Primary HDU metadata
        self.psf_pitch         = 3.0 
        self.psf_width_pix     = psfs_fits[1].data.shape[0]
        self.psf_width_as      = self.psf_width_pix * self.pixsize
        
        self.fourier_tile_dim = self.gauss_dim + np.array(psfs_fits[1].data.shape)
        
        if which_psf is None:
            # Use variable PSFs
            self.static = False
            # Create PSF objects and do their FFTs
            print("Using varible PSFs")
            print("Doing PSF FFTs:")
            self.psfs = []
            for psf in tqdm(psfs_fits[1:],leave=False):
                self.psfs.append(PSF(psf, self.fourier_tile_dim))
            psfs_fits.close()
            print("Done.                          ")
        else:
            # Use static PSFs
            self.static = True
            self.psfs = []
            self.psfs.append(PSF(psfs_fits[1+which_psf], self.fourier_tile_dim))
            psfs_fits.close()
            print(f"Using static PSF: {which_psf:3d} (fits hdu {which_psf+1:3d})" + 
                    f"\nPSF pos: ({self.psfs[0].xpos:0.3f},{self.psfs[0].ypos:0.3f})\"")

        self._init = True
        
    def get_effective_psf_fft(self, s_pos):
        """Takes star information and computes effective PSF.

        From star position, the convex combination PSF is found
        (equivalent to bilinear interpolation since stars are defined
        on square grid). The resulting effective PSF is added to the 
        internal _psf_array to be used in the get_tile pipeline.

        Args:
            star_pos (`np.ndarray`) : position [arcsec].
        """
        
        if self.static == True:
            # Use static PSF
            self._psf_array += self.psfs[0].fft_data
        else:
            # Perform convex combination of PSFs based on star position
            for psf in self.psfs:
                gamma_check = \
                    (np.abs(psf.xpos-s_pos[0]) < self.psf_pitch) and \
                    (np.abs(psf.ypos-s_pos[1]) < self.psf_pitch)
                if gamma_check:
                    # PSF is a vertex of convex hull around star.
                    # Compute the PSFs convex combination coefficient:
                    gamma = (
                                (1-np.abs(psf.xpos-s_pos[0])/self.psf_pitch) * \
                                (1-np.abs(psf.ypos-s_pos[1])/self.psf_pitch)
                             ).astype(psf.fft_data.dtype)
                else:
                    continue
                # add result to PSF array
                self._psf_array += gamma*psf.fft_data
    
    def get_tile(self,index):
        """Get the tile corresponding to source[index]

        From the tile_generator object tgen, calling tgen.get_tile(index) will
        generate the tile corresponding to the tgen.source_pos[index] star by 
        interpolating the 4 neighbouring PSFs and convolving this effective
        PSF with a sub-pixel shifted Dirac-delta function, and if requested, a 
        Gaussian kernel defined by tgen.cov_mat .

        The output of this is a tile which has been trimmed down to the input
        PSF dimensions, as well as the coordinates of the bottom-left-corner
        of the tile so that it may be sliced into the final image properly.
        
        Args:
            index (`int`): index of star in source table to generate tile for.

        Returns:
            out (real-valued `np.ndarray`): tile to be sliced into final image
            bottom_left_corner (`np.ndarray`): coordinates of bottom left corner of bottom left pixel in arcsec
        """
        
        if self._init == True:
            # Initialise internal variables first time around
            
            # new shape is ~ half full tile size due to RFFT2 optimisation
            new_shape = [self.fourier_tile_dim[0],self.fourier_tile_dim[1]//2+1]
            self._psf_array = np.zeros(new_shape,dtype=self.dtype)

            # Coordinates (uu,vv) of DFT space:
            self._nx = self.fourier_tile_dim[0] # square image only
            self._T  = self.dtype(self.gauss_width_as + self.psf_width_as)
            self._Ts = self.dtype(self._T/self._nx)
            uu,vv = np.meshgrid(
                np.linspace(0,1/self._Ts,self._nx+1,dtype=self.dtype)[:-1]-1/(2*self._Ts),
                np.linspace(0,1/self._Ts,self._nx+1,dtype=self.dtype)[:-1]-1/(2*self._Ts)
                )
            uu = np.fft.fftshift(uu)
            vv = np.fft.fftshift(vv)
            self._fft_pos = np.c_[uu[:new_shape[0],:new_shape[1]].flatten(),
                                 vv[:new_shape[0],:new_shape[1]].flatten()].T
            
            # Run einsum calcs once to determine optimal sequence:
            self.optimize_star_kernel()
            
            self._init = False
        
        # Pick star from source table
        star_flux = self.source_flux[index]
        star_pos  = self.source_pos[index] - 0.5*self.pixsize # convention is that (0,0)" is between the middle 4 pixels"
        
        if self.dtype==np.complex64:
            star_flux = np.float32(star_flux)
            star_pos  = star_pos.astype(np.float32)

        # Clear internal array:
        self._psf_array *= 0.0

        # Compute effective PSF:
        self.get_effective_psf_fft(star_pos)

        # Prepare for FFT Gaussian computation:
        offset = star_pos%self.pixsize
        _star_pos = (self.psf_width_as+self.gauss_width_as)/2 * np.r_[1.0,1.0] + offset
        bottom_left_corner = np.int32(star_pos//self.pixsize - self.psf_width_pix//2)

        # Compute star Gaussian and convolve with PSF:
        self._psf_array *= self.get_star_kernel_fft(star_flux, _star_pos)
        
        # Inverse FFT and trimming:
        out = (np.fft.fftshift(
            np.fft.irfft2(
                (self._psf_array)
            ).astype(self.dtype))).real[:self.psf_width_pix,:self.psf_width_pix]

        return out, bottom_left_corner    

    def get_star_kernel_fft(self, flux, mu):
        """Compute star Gaussian based in DFT space.

        Directly computes the FFT of the Gaussian kernel with appriate amplitude, 
        width, and offset to suit the tile being generated.

        Uses optimised np.einsum so requires running `optimize_star_kernel()` first.

        Args:
            flux (`float`): flux of star.
            mu (`np.ndarray`): position of star.

        Returns:
            gaussian_fft (`np.ndarray`): star Gaussian kernel in FFT space.
        """
        offset = (2*np.pi*1j)*mu
        if self.cov_mat is None:
            gaussian_fft = flux*np.exp(
                -np.einsum(
                    "ij,i->j",
                    self._fft_pos,
                    offset,
                    optimize=self._esp2
                )).reshape(self._psf_array.shape) 
        else:
            gaussian_fft = flux*np.exp(
                (-2*(np.pi)**2)*np.einsum(
                    "ij,ii,ij->j",
                    self._fft_pos,
                    self.cov_mat,
                    self._fft_pos,
                    optimize=self._esp1
                ) - np.einsum(
                    "ij,i->j",
                    self._fft_pos,
                    offset,
                    optimize=self._esp2
                )).reshape(self._psf_array.shape) 
        return gaussian_fft
    
    def optimize_star_kernel(self):
        """Runs star kernel once to optimise `np.einsum`
        """

        cov = np.zeros([2,2],dtype=np.float32 if self.dtype == np.complex64 else np.float64)
        mu  = self.source_pos[0]
        offset = 2*np.pi*1j*mu.flatten()
        self._esp1 = np.einsum_path(
                    "ij,ii,ij->j",
                    self._fft_pos,
                    cov,
                    self._fft_pos,
                    optimize="optimal"
                )[0]
        self._esp2 = np.einsum_path(
                    "ij,i->j",
                    self._fft_pos,
                    -offset,
                    optimize="optimal"
                )[0]
    

class ImageGenerator:
    """Generate image from sliced tiles, one tile per source object.

    This is the core object to work with when generating a MAVISIM image from
    a `Source` object.

    Args:
        array_width_pix (`int`): width of full image in pixels before rebinning.
        pixsize (`float`): pixel size in arcsec before rebinning.
        source (`Source` object): source list as `Source`-type object
        psfs_file (`str`): filename for fits file containing all PSFs and PSF metadata.
        gauss_width_pix (`int`): width of Gaussian square support in pixels.
        which_psf (`int`, optional): If specified, use fits HDU[which_psf+1] PSF only.
    
    Attributes:
        pixsize (`float`): pixel size in arcsec of image.
        fov (`float`): FoV of full image.
        full_image (real-valued `np.ndarray`): final image at original pixel size (i.e., before rebinning).
        tile_gen (`TileGenerator` object): tile generator object used to create tiles to slice into final image.
    """
    def __init__(self, array_width_pix, source, psfs_file, pixsize=3.75e-3, gauss_width_pix=34,
            which_psf = None):        
        """Contructor for ImageGenerator object
        """

        self.pixsize = pixsize
        self.xx = np.arange(array_width_pix)*pixsize
        self.fov = self.xx[-1]-self.xx[0]
        self.xx -= (self.fov/2+self.pixsize/2)
        self.full_image = np.zeros([self.xx.shape[0],self.xx.shape[0]])
        self.tile_gen = TileGenerator(source, psfs_file, gauss_width_pix, which_psf=which_psf)
        self.nsource = source.flux.shape[0]
        
    def main(self):
        """Loop over all stars and add the tile to the full image.
        """
        for ni in tqdm(range(self.nsource),leave=False):
            # Generate the tile:
            tile,origin = self.tile_gen.get_tile(ni)
            # Find the location of the tile in the full image:
            xstart = origin[0]+self.xx.shape[0]//2
            ystart = origin[1]+self.xx.shape[0]//2
            # Slice the tile in:
            self.full_image[ystart:ystart+self.tile_gen.psf_width_pix,
                            xstart:xstart+self.tile_gen.psf_width_pix] += tile
        
    def get_rebinned_cropped(self,rebin_factor,cropped_width_as):
        """Rebin self.full_image after cropping to desired rebin factor.
        
        Args:
            rebin_factor (int): rebinning factor from high-res image to rebinned image. 
            Note that no checking is done on the validity of this, so use with care.
            cropped_width_as (float): desired width of final image in arcsec.
        Returns:
            rebinned_im (real-valued `np.ndarray`): complete image, rebinned and cropped. 
        """

        xx_cropped_id = np.abs(self.xx)<=cropped_width_as/2
        cropped_im = self.full_image[xx_cropped_id,:][:,xx_cropped_id]
        rebinned_im = self._rebin(cropped_im,np.array(cropped_im.shape)//rebin_factor)
        return rebinned_im

    def _rebin(self, arr, new_shape):
        """Rebin array (arr) into new shape (new_shape)."""
        shape = (new_shape[0], arr.shape[0] // new_shape[0],
                 new_shape[1], arr.shape[1] // new_shape[1])
        return arr.reshape(shape).sum(-1).sum(1)

